// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.java.visitor;

import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.NonTerminalProgramElement;

class PosTracker {

    /**
     * reference to the next element in stack
     */
    private PosTracker parent;

    /**
     * 
     */
    private ProgramElement pe;

    /**
     * max. amount of children
     */
    private int childCount;

    /**
     * the current visited child
     */
    private int currentChildPos;
	
	
    /** 
     * the program element and the number of 
     * the child that is visited 
     */
    private PosTracker(ProgramElement pe, 
		       int currentChildPos) {
	this.pe = pe;
	childCount = pe instanceof NonTerminalProgramElement ? 
	    ((NonTerminalProgramElement)pe).getChildCount() : 0;
	this.currentChildPos = currentChildPos;
    }


    /** 
     * the program element and the number of 
     * the child that is visited 
     */
    private PosTracker(NonTerminalProgramElement pe, 
		       int currentChildPos) {
	this.pe = pe;
	childCount = pe.getChildCount();
	this.currentChildPos = currentChildPos;
    }

    private void setParent(PosTracker parent) {
	this.parent = parent;
    }
    private ProgramElement node() {
	return pe;
    }

    private PosTracker parent() {
	return parent;
    }

    private int pos() {
	return currentChildPos;
    }

    private void setLast() {
	if (pe instanceof NonTerminalProgramElement) {
	    currentChildPos = childCount;
	} else {
	    currentChildPos = -1;
	}
    }

    private int getChildCount() {
	return childCount;
    }

    private boolean hasNext() {
	return currentChildPos < childCount;
    }
	
    private void setPos(int n) {
	currentChildPos = n;
    }

    private void incPos() {
	currentChildPos++;
    }

    private void decPos() {
	currentChildPos--;
    }
	
    public String toString() {
	return "PE:"+pe+"|| Pos:"+currentChildPos;
    }

}   
